/* eslint-disable capitalized-comments */
/* eslint-env mocha */
'use strict';

const assert = require('assert');
const sinon = require('sinon'); // eslint-disable-line node/no-unpublished-require
const Router = require('../../lib/router');
const PathParameter = require('../../lib/router/path-parameter')
const { DingusError } = require('../../lib/errors');

const noExpectedException = 'did not get expected exception';

describe('Router', function () {
  const router = new Router();
  let _its; // Save and restore ignoreTrailingSlash

  beforeEach(function () {
    _its = router.ignoreTrailingSlash;
  });

  afterEach(function () {
    sinon.restore();
    router.ignoreTrailingSlash = _its;
  });

  describe('_pathToRoutePath', function () {
    it('defines a simple path', function () {
      const p = '/a/b/c';
      const expected = ['', 'a', 'b', 'c'];
      expected[Router.kPathMethods] = {};
      const r = router._pathToRoutePath(p);
      assert.deepStrictEqual(r, expected);
    });
    it('defines a path with parameter', function () {
      const p = '/a/b/:c';
      const expected = ['', 'a', 'b', new PathParameter('c')];
      expected[Router.kPathMethods] = {};
      const r = router._pathToRoutePath(p);
      assert.deepStrictEqual(r, expected);
    });
    it('defines a path with trailing slash', function () {
      router.ignoreTrailingSlash = true;
      const p = '/a/b/:c/';
      const expected = ['', 'a', 'b', new PathParameter('c')];
      expected[Router.kPathMethods] = {};
      const r = router._pathToRoutePath(p);
      assert.deepStrictEqual(r, expected);
    });
  }); // _pathToRoutePath

  describe('_pathCompareExact', function () {
    let fixedPath, checkPath;

    it('compares static paths which match', function () {
      fixedPath = router._pathToRoutePath('/a/b/c');
      checkPath = router._pathToRoutePath('/a/b/c');
      const r = Router._pathCompareExact(fixedPath, checkPath);
      assert.strictEqual(r, true);
    });
    it('compares static paths which do not match', function () {
      fixedPath = router._pathToRoutePath('/a/b/c');
      checkPath = router._pathToRoutePath('/a/b/d');
      const r = Router._pathCompareExact(fixedPath, checkPath);
      assert.strictEqual(r, false);
    });
    it('compares unequal static paths', function () {
      fixedPath = router._pathToRoutePath('/a/b/c');
      checkPath = router._pathToRoutePath('/a/b');
      const r = Router._pathCompareExact(fixedPath, checkPath);
      assert.strictEqual(r, false);
    });
    it('compares param paths which match', function () {
      fixedPath = router._pathToRoutePath('/a/:b/c');
      checkPath = router._pathToRoutePath('/a/:b/c');
      const r = Router._pathCompareExact(fixedPath, checkPath);
      assert.strictEqual(r, true);
    });
    it('compares param paths which do not match', function () {
      fixedPath = router._pathToRoutePath('/a/:b/c');
      checkPath = router._pathToRoutePath('/a/:g/c');
      const r = Router._pathCompareExact(fixedPath, checkPath);
      assert.strictEqual(r, false);
    });
  }); // _pathCompareExact

  describe('_pathCompareParam', function () {
    let fixedPath, checkPath;

    it('compares static paths which match', function () {
      const params = {};
      const expectedParams = {};
      fixedPath = router._pathToRoutePath('/a/b/c');
      checkPath = router._pathToRoutePath('/a/b/c');
      const r = Router._pathCompareParam(fixedPath, checkPath);
      assert.strictEqual(r, true);
      assert.deepStrictEqual(params, expectedParams);
    });
    it('compares static paths which do not match', function () {
      const params = {};
      const expectedParams = {};
      fixedPath = router._pathToRoutePath('/a/b/c');
      checkPath = router._pathToRoutePath('/a/b/d');
      const r = Router._pathCompareParam(fixedPath, checkPath, params);
      assert.strictEqual(r, false);
      assert.deepStrictEqual(params, expectedParams);
    });
    it('compares unequal static paths', function () {
      const params = {};
      const expectedParams = {};
      fixedPath = router._pathToRoutePath('/a/b/c');
      checkPath = router._pathToRoutePath('/a/b');
      const r = Router._pathCompareParam(fixedPath, checkPath, params);
      assert.strictEqual(r, false);
      assert.deepStrictEqual(params, expectedParams);
    });
    it('compares param paths which match', function () {
      const params = {};
      const expectedParams = {
        b: 'bar',
      };
      fixedPath = router._pathToRoutePath('/a/:b/c');
      checkPath = router._pathToRoutePath('/a/bar/c');
      const r = Router._pathCompareParam(fixedPath, checkPath, params);
      assert.strictEqual(r, true);
      assert.deepStrictEqual(params, expectedParams);
    });
    it('compares multi param path which match', function () {
      const params = {};
      const expectedParams = {
        b: 'gaz',
        c: '123',
      };
      fixedPath = router._pathToRoutePath('/a/:b/:c');
      checkPath = router._pathToRoutePath('/a/gaz/123');
      const r = Router._pathCompareParam(fixedPath, checkPath, params);
      assert.strictEqual(r, true);
      assert.deepStrictEqual(params, expectedParams);
    });
  }); // _pathCompareParam

  describe('_pathFind / _pathFindExact', function () {
    let pathsByLengthOrig;

    beforeEach(function () {
      pathsByLengthOrig = router.pathsByLength;
      router.pathsByLength = {
        2: [ router._pathToRoutePath('/:id') ],
        3: [ router._pathToRoutePath('/a/b') ],
      };
    });
    afterEach(function () {
      router.pathsByLength = pathsByLengthOrig;
    });
  
    describe('_pathFind', function () {
      it('finds a path', function () {
        const pathParts = ['', '46123f1e-bdca-40ee-9f62-93ad38647aa1'];
        const { matchedPath, pathParams } = router._pathFind(pathParts);
        assert.strictEqual(matchedPath, router.pathsByLength[2][0]);
        assert.deepStrictEqual(pathParams, { id: pathParts[1] });
      });
      it('does not find a path', function () {
        const pathParts = ['', 'flarp', 'baz'];
        const { matchedPath } = router._pathFind(pathParts);
        assert.strictEqual(matchedPath, undefined);
      });
    }); // _pathFind

    describe('_pathFindExact', function () {
      it('finds a path', function () {
        const pathParts = ['', new PathParameter('id')];
        const r = router._pathFindExact(pathParts);
        assert.strictEqual(r, router.pathsByLength[2][0]);
      });
      it('does not find a path', function () {
        const pathParts = ['', 'flarp', 'baz'];
        const r = router._pathFindExact(pathParts);
        assert.strictEqual(r, undefined);
      });
    }); // _pathFindExact

  }); // _pathFind / _pathFindExact

  describe('on', function () {
    let pathsByLengthOrig;
    const stubHandler = () => {};
    const stubEntry = {
      handler: stubHandler,
      handlerArgs: [],
    };

    beforeEach(function () {
      pathsByLengthOrig = router.pathsByLength;
      router.pathsByLength = {
        2: [ router._pathToRoutePath('/:id') ],
      };
    });
    afterEach(function () {
      router.pathsByLength = pathsByLengthOrig;
    });
  
    it('adds new path', function () {
      const urlPath = '/a/:id';
      const expected = router._pathToRoutePath(urlPath);
      expected[Router.kPathMethods]['GET'] = stubEntry;
      router.on('GET', urlPath, stubHandler);
      assert.deepStrictEqual(router.pathsByLength[3][0], expected);
    });

    it('adds new method to path', function () {
      const urlPath = '/a/:id';
      const expected = router._pathToRoutePath(urlPath);
      expected[Router.kPathMethods]['GET'] = stubEntry;
      expected[Router.kPathMethods]['POST'] = stubEntry;
      router.on('GET', urlPath, stubHandler);

      router.on('POST', urlPath, stubHandler);
      assert.deepStrictEqual(router.pathsByLength[3][0], expected);
    });

    it('add some more paths', function () {
      let urlPath = '/a/b/c/d';
      const expected = router._pathToRoutePath(urlPath);
      expected[Router.kPathMethods]['GET'] = stubEntry;
      router.on('GET', urlPath, stubHandler);
      urlPath = '/a/b/x/y';
      router.on('GET', urlPath, stubHandler);

      assert.strictEqual(router.pathsByLength[5].length, 2);
    });

    it('adds multiple methods', function () {
      const urlPath = '/:id';
      const expected = router._pathToRoutePath(urlPath);
      expected[Router.kPathMethods]['GET'] = stubEntry;
      expected[Router.kPathMethods]['HEAD'] = stubEntry;

      router.on(['GET', 'HEAD'], urlPath, stubHandler);
      assert.deepStrictEqual(router.pathsByLength[2][0], expected);
    });

    it('adds new wildcard path', function () {
      const urlPath = '/a/:id';
      const expected = router._pathToRoutePath(urlPath);
      expected[Router.kPathMethods]['*'] = stubEntry;
      router.on('*', urlPath, stubHandler);
      assert.deepStrictEqual(router.pathsByLength[3][0], expected);
    });

    it('fails to add unknown method path', function () {
      const urlPath = '/a/:id';
      try {
        router.on('FLARP', urlPath, stubHandler);
        assert.fail('expected an exception');
      } catch (e) {
        assert.strictEqual(e.name, 'DingusError');
        assert.strictEqual(e.message, 'invalid method \'FLARP\'');
      }
    });

    it('requires args to be array', function () {
      const urlPath = '/a';
      try {
        router.on('GET', urlPath, stubHandler, {});
        assert.fail('expected an exception');
      } catch (e) {
        assert(e instanceof TypeError);
      }
    });
  }); // on

  describe('lookup', function () {
    let pathsByLengthOrig;
    let ctx;
    let stubHandler;

    beforeEach(function () {
      ctx = {};
      pathsByLengthOrig = router.pathsByLength;
      stubHandler = sinon.stub();
    });
    afterEach(function () {
      router.pathsByLength = pathsByLengthOrig;
    });

    it('finds handler', function () {
      const urlPath = '/:id';
      const method = 'GET';
      router.on(method, urlPath, stubHandler);
      const path = '/abc';

      const { handler } = router.lookup(method, path, ctx);
      assert.strictEqual(handler, stubHandler);
    });
    it('does not find handler with trailing slash', function () {
      router.ignoreTrailingSlash = false;
      const urlPath = '/:id';
      const method = 'GET';
      router.on(method, urlPath, stubHandler);
      const path = '/abc/';

      try {
        router.lookup(method, path, ctx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert(e instanceof DingusError);
        assert.strictEqual(e.message, 'NoPath');
      }
    });
    it('finds handler', function () {
      router.ignoreTrailingSlash = true;
      const urlPath = '/:id';
      const method = 'GET';
      router.on(method, urlPath, stubHandler);
      const path = '/abc/';

      const { handler } = router.lookup(method, path, ctx);
      assert.strictEqual(handler, stubHandler);
    });
    it('finds handler without context', async function () {
      const urlPath = '/:id';
      const method = 'GET';
      router.on(method, urlPath, stubHandler);
      const path = '/abc';

      const { handler } = router.lookup(method, path);
      assert.strictEqual(handler, stubHandler);
    });
    it('finds fallback handler', async function () {
      const urlPath = '/abc/:id';
      const method = 'GET';
      router.on('*', urlPath, stubHandler);
      const path = '/abc/def';

      const { handler } = router.lookup(method, path, ctx);
      assert.strictEqual(handler, stubHandler);
    });
    it('calls unsupported method', async function () {
      const urlPath = '/:id';
      const method = 'POST';
      router.on('GET', urlPath, stubHandler);
      const path = '/abc';

      try {
        router.lookup(method, path, ctx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert(e instanceof DingusError);
        assert.strictEqual(e.message, 'NoMethod');
      }
    });
    it('does not lookup non-existent path', async function () {
      const path = '/foo/bar';
      const method = 'GET';

      try {
        router.lookup(method, path, ctx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert(e instanceof DingusError);
        assert.strictEqual(e.message, 'NoPath');
      }
    });

  }); // lookup

});
